/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001  Ludvig Strigeus
 * Copyright (C) 2001-2005 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /cvsroot/scummvm/scummvm/common/scaler/2xsai.cpp,v 1.9 2005/07/30 21:10:58 sev Exp $
 *
 */

#include "common/scaler/intern.h"


#undef PI
#define SHOW_CHANGES 0
//#define ALPHA_BLEND_128_W(dst, src) dst = ((src & highBits) >> 1) + ((dst & highBits) >> 1)
#define ALPHA_BLEND_128_W(dst, src) dst = INTERPOLATE<bitFormat>(dst, src)


template<int bitFormat>
void _2xPM_LQTemplate(const uint8 *src, uint32 src_pitch, uint8 *dest, uint32 dstPitch, int src_width, int src_height) {
	unsigned long x, y;
	unsigned short int PA, PB, PC, PD, PE, PF, PG, PH, PI;
	register const unsigned short int *start_addr1, *start_addr2, *start_addr3;
	unsigned long next_line, next_line_src;
	unsigned short int *dst_pixel;
	unsigned long complete_line_src, complete_line_dst;
	unsigned char auto_blend;
	unsigned short int E[4];
	unsigned char dont_reblit;

	next_line_src = src_pitch >> 1;
	next_line = dstPitch >> 1;

	complete_line_src = next_line_src - src_width;
	complete_line_dst = dstPitch - (src_width << 1);

	start_addr2 = (const unsigned short int *)(src - 2);

	dst_pixel = (unsigned short int *)(dest);

	y = src_height - 1;
	//for (y = 0; y < src_height; y++)
	do
	{
		start_addr1 = start_addr2 - next_line_src;
		start_addr3 = start_addr2 + next_line_src;

		auto_blend = 0;
		x = src_width - 1;

		//for (x = 0; x < src_width; x++)
		do
		{
			PA = start_addr1[0];
			PB = start_addr1[1];
			PC = start_addr1[2];

			PD = start_addr2[0];
			PE = start_addr2[1];
			PF = start_addr2[2];

			PG = start_addr3[0];
			PH = start_addr3[1];
			PI = start_addr3[2];

#if SHOW_CHANGES
			unsigned short int PDE = PE;
			PE = 0;
#endif
			E[0] = E[1] = E[2] = E[3] = PE;

			dont_reblit = 0;

			// Horizontal
			if (!dont_reblit)
			if (PD != PF)
			{
				if ((PE != PD) && (PD == PH) && (PD == PI) && (PE != PG)
					&& ((PD != PG) || (PE != PF) || (PA != PD))
					&& (!((PD == PA) && (PD == PG) && (PE == PB) && (PE == PF)))
					)
				{
					E[2] = PH;
					ALPHA_BLEND_128_W(E[3], PH);
					dont_reblit = 1;
				}

				else if ((PE != PF) && (PF == PH) && (PF == PG) && (PE != PI)
					&& ((PF != PI) || (PE != PD) || (PC != PF))
					&& (!((PF == PC) && (PF == PI) && (PE == PB) && (PE == PD)))
					)
				{
					ALPHA_BLEND_128_W(E[2], PH);
					E[3] = PH;
					dont_reblit = 1;
				}
			}

			if (!dont_reblit)
			{
				if ((PB != PH) && (PD != PF))
				{
					if ((PB == PD) && (PE != PD)
						&& (!((PE == PF) && (((PE == PA) && (PB == PC))
							             || ((PD == PA) && (PD == PC) && (PG != PD) && (PG != PE)))))
						&& (!((PB == PA) && (PB == PG))))
						ALPHA_BLEND_128_W(E[0], PB);

					else if ((PB == PF) && (PE != PF)
						&& (!((PE == PD) && (((PE == PC) && (PB == PA))
						                 || ((PF == PA) && (PF == PC) && (PE == PD) && (PI != PF) && (PI != PE)))))
						&& (!((PB == PC) && (PB == PI)))
						)
						ALPHA_BLEND_128_W(E[1], PB);

					if ((PH == PD) && ((PE != PG) || (PE != PD) || ((PE != PF) && (PF == PI)))
						&& (!((PE == PF) && (((PE == PG) && (PH == PI))
						                 || ((PD == PG) && (PD == PI) && (PE == PF) && (PA != PD) && (PA != PE))
										 || ((PE == PG) && (PH == PI)))))
						&& (!((PH == PG) && (PH == PA)))
						)
						ALPHA_BLEND_128_W(E[2], PH);

					else if ((PH == PF) && ((PE != PI) || (PE != PF))
						&& (!((PE == PD) && (((PE == PI) && (PH == PG))
						                 || ((PF == PG) && (PF == PI) && (PC != PF) && (PI != PE)))))
						&& (!((PH == PI) && (PH == PC)))
						)
						ALPHA_BLEND_128_W(E[3], PH);
				} else
#if SHOW_CHANGES
				PE = PDE;
#endif
				if ((PD == PB) && (PD == PF) && (PD == PH) && (PD != PE))
				{
					if ((PD == PG) || (PD == PC))
					{
						ALPHA_BLEND_128_W(E[1], PD);
						E[2] = E[1];
					}
					if ((PD == PA) || (PD == PI))
					{
						ALPHA_BLEND_128_W(E[0], PD);
						E[3] = E[0];
					}
				}
			}

			dst_pixel[0] = E[0];
			dst_pixel[1] = E[1];
			dst_pixel[next_line] = E[2];
			dst_pixel[next_line + 1] = E[3];

			start_addr1++;
			start_addr2++;
			start_addr3++;

			dst_pixel += 2;
		} while (x--);

		start_addr2 += complete_line_src;
		dst_pixel += complete_line_dst;
	} while(y--);
}

MAKE_WRAPPER(_2xPM_LQ)
